@@ -6,6 +6,7 @@
 #include <sys/types.h>
 #include <net/if.h>
 #include <stdint.h>
+#include <stdarg.h>
 
 #define E_syntax	  2
 #define E_usage		  3
@@ -25,6 +26,8 @@ enum {
 
   SUPRESS_STDERR       = 16,
 
+  RETURN_STDOUT_FD     = 32,
+
   DONT_REPORT_FAILED   = 0x80,
 };
 
@@ -63,28 +66,39 @@ struct ifi_info {
   struct ifi_info *ifi_next;    /* next ifi_info structure */
 };
 
+struct d_name
+{
+  char *name;
+  struct d_name *next;
+};
+
 struct d_proxy_info
 {
-  char* name;
+  struct d_name *on_hosts;
   char* inside_addr;
   char* inside_port;
+  char* inside_af;
   char* outside_addr;
   char* outside_port;
+  char* outside_af;
 };
 
 struct d_host_info
 {
-  char* name;
+  struct d_name *on_hosts;
   char* device;
   char* disk;
   char* address;
   char* port;
   char* meta_disk;
+  char* address_family;
   int meta_major;
   int meta_minor;
   char* meta_index;
   struct d_proxy_info *proxy;
   struct d_host_info* next;
+  struct d_resource* lower;  /* for device stacking */
+  int config_line;
 };
 
 struct d_option
@@ -94,16 +108,20 @@ struct d_option
   struct d_option* next;
   unsigned int mentioned  :1 ; // for the adjust command.
   unsigned int is_default :1 ; // for the adjust command.
+  unsigned int is_escaped :1 ;
 };
 
 struct d_resource
 {
   char* name;
   char* protocol;
-  char* device;     // gets propagated to host_info sections later.
-  char* disk;       // gets propagated to host_info sections later.
-  char* meta_disk;  // gets propagated to host_info sections later.
-  char* meta_index; // gets propagated to host_info sections later.
+
+  /* these get propagated to host_info sections later. */
+  char* device;
+  char* disk;
+  char* meta_disk;
+  char* meta_index;
+
   struct d_host_info* me;
   struct d_host_info* peer;
   struct d_host_info* all_hosts;
@@ -114,6 +132,11 @@ struct d_resource
   struct d_option* handlers;
   struct d_option* proxy_options;
   struct d_resource* next;
+  struct d_name *become_primary_on;
+  unsigned int stacked_timeouts:1;
+  unsigned int ignore:1;
+  unsigned int stacked:1;
+  int me_minor; /* cache local minor number of this resource */
 };
 
 extern int adm_attach(struct d_resource* ,const char* );
@@ -125,6 +148,7 @@ extern int _admm_generic(struct d_resource* ,const char*, int flags);
 extern int m_system(char**, int, struct d_resource*);
 extern struct d_option* find_opt(struct d_option*,char*);
 extern void validate_resource(struct d_resource *);
+extern int vcheck_uniq(const char* what, const char *fmt, va_list ap);
 extern int check_uniq(const char* what, const char *fmt, ...);
 extern void verify_ips(struct d_resource* res);
 extern void schedule_dcmd( int (* function)(struct d_resource*,const char* ),
@@ -137,6 +161,11 @@ extern int adm_create_md(struct d_resource* res ,const char* cmd);
 extern void convert_discard_opt(struct d_resource* res);
 extern void convert_after_option(struct d_resource* res);
 
+/* See drbdadm_minor_table.c */
+extern int register_minor(int minor, const char *path);
+extern int unregister_minor(int minor);
+extern char *lookup_minor(int minor);
+
 enum pr_flags {
   ThisHRequired = 1,
   PeerHRequired = 2,
@@ -145,6 +174,13 @@ enum pr_flags {
 };
 extern struct d_resource* parse_resource(char*, enum pr_flags);
 extern struct d_option *new_opt(char *name, char *value);
+extern int name_in_names(char *name, struct d_name *names);
+extern char *_names_to_str(char* buffer, struct d_name *names);
+extern char *_names_to_str_c(char* buffer, struct d_name *names, char c);
+#define NAMES_STR_SIZE 255
+#define names_to_str(N) _names_to_str(alloca(NAMES_STR_SIZE+1), N)
+#define names_to_str_c(N, C) _names_to_str_c(alloca(NAMES_STR_SIZE+1), N, C)
+void free_names(struct d_name *names);
 
 extern char* config_file;
 extern int config_valid;
@@ -187,3 +223,16 @@ extern int soi;
 	for (res = (config); res && (tmp = res->next, 1); res = tmp)
 
 #endif
+
+#define APPEND(LIST,ITEM) ({		      \
+  typeof((LIST)) _l = (LIST);		      \
+  typeof((ITEM)) _i = (ITEM);		      \
+  typeof((ITEM)) _t;			      \
+  _i->next = NULL;			      \
+  if (_l == NULL) { _l = _i; }		      \
+  else {				      \
+    for (_t = _l; _t->next; _t = _t->next);   \
+    _t->next = _i;			      \
+  };					      \
+  _l;					      \
+})
