@@ -38,21 +38,86 @@ YYSTYPE yylval;
 
 /////////////////////
 
-#define APPEND(LIST,ITEM) ({		      \
-  typeof((LIST)) _l = (LIST);		      \
-  typeof((ITEM)) _i = (ITEM);		      \
-  typeof((ITEM)) _t;			      \
-  _i->next = NULL;			      \
-  if (_l == NULL) { _l = _i; }		      \
-  else {				      \
-    for (_t = _l; _t->next; _t = _t->next);   \
-    _t->next = _i;			      \
-  };					      \
-  _l;					      \
-})
-
 static int c_section_start;
 
+struct d_name *names_from_str(char* str)
+{
+	struct d_name *names;
+
+	names = malloc(sizeof(struct d_name));
+	names->next = NULL;
+	names->name = strdup(str);
+
+	return names;
+}
+
+char *_names_to_str_c(char* buffer, struct d_name *names, char c)
+{
+	int n = 0;
+
+	while (1) {
+		n += snprintf(buffer + n, NAMES_STR_SIZE - n, "%s", names->name);
+		names = names->next;
+		if (!names)
+			return buffer;
+		n += snprintf(buffer + n, NAMES_STR_SIZE - n, "%c", c);
+	}
+}
+
+char *_names_to_str(char* buffer, struct d_name *names)
+{
+	return _names_to_str_c(buffer, names, ' ');
+}
+
+int name_in_names(char *name, struct d_name *names)
+{
+	while (names) {
+		if (!strcmp(names->name, name))
+			return 1;
+		names = names->next;
+	}
+	return 0;
+}
+
+void free_names(struct d_name *names)
+{
+	struct d_name *nf;
+	while (names) {
+		nf = names->next;
+		free(names->name);
+		free(names);
+		names = nf;
+	}
+}
+
+static void append_names(struct d_name **head, struct d_name ***last, struct d_name *to_copy)
+{
+	struct d_name *new;
+
+	while (to_copy) {
+		new = malloc(sizeof(struct d_name));
+		if (!*head)
+			*head = new;
+		new->name = strdup(to_copy->name);
+		new->next = NULL;
+		if (*last)
+			**last = new;
+		*last = &new->next;
+		to_copy = to_copy->next;
+	}
+}
+
+
+struct d_name *concat_names(struct d_name *to_copy1, struct d_name *to_copy2)
+{
+	struct d_name *head = NULL, **last = NULL;
+
+	append_names(&head, &last, to_copy1);
+	append_names(&head, &last, to_copy2);
+
+	return head;
+}
+
 void m_strtoll_range(const char *s, char def_unit,
 		     const char *name,
 		     unsigned long long min, unsigned long long max)
@@ -92,9 +157,9 @@ void range_check(const enum range_checks what, const char *name,
 				DRBD_DIALOG_REFRESH_MAX);
 		break;
 	case R_DISK_SIZE:
-		m_strtoll_range(value, 'K', name,
-				DRBD_DISK_SIZE_SECT_MIN >> 1,
-				DRBD_DISK_SIZE_SECT_MAX >> 1);
+		m_strtoll_range(value, 's', name,
+				DRBD_DISK_SIZE_SECT_MIN,
+				DRBD_DISK_SIZE_SECT_MAX);
 		break;
 	case R_TIMEOUT:
 		m_strtoll_range(value, 1, name, DRBD_TIMEOUT_MIN,
@@ -159,6 +224,7 @@ struct d_option *new_opt(char *name, char *value)
 	cn->value = value;
 	cn->mentioned = 0;
 	cn->is_default = 0;
+	cn->is_escaped = 0;
 
 	return cn;
 }
@@ -167,11 +233,33 @@ static void derror(struct d_host_info *host, struct d_resource *res, char *text)
 	config_valid = 0;
 	fprintf(stderr, "%s:%d: in resource %s, on %s { ... }:"
 		" '%s' keyword missing.\n",
-		config_file, c_section_start, res->name, host->name, text);
+		config_file, c_section_start, res->name, names_to_str(host->on_hosts), text);
 }
 
+static void pperror(struct d_host_info *host, struct d_proxy_info *proxy, char *text)
+{
+	config_valid = 0;
+	fprintf(stderr, "%s:%d: in section: on %s { proxy on %s { ... } }:"
+		" '%s' keyword missing.\n",
+		config_file, c_section_start, names_to_str(host->on_hosts),
+		names_to_str(proxy->on_hosts), text);
+}
+
+#define typecheck(type,x) \
+({	type __dummy; \
+	typeof(x) __dummy2; \
+	(void)(&__dummy == &__dummy2); \
+	1; \
+})
+
+#define for_each_host(h_,hosts_) \
+	for ( ({ typecheck(struct d_name*, h_); \
+		h_ = hosts_; }); \
+	 	h_; h_ = h_->next)
+
 void check_meta_disk(struct d_host_info *host)
 {
+	struct d_name *h;
 	if (strcmp(host->meta_disk, "internal") != 0) {
 		/* external */
 		if (host->meta_index == NULL) {
@@ -180,8 +268,8 @@ void check_meta_disk(struct d_host_info *host)
 				config_file, fline, host->meta_disk);
 		}
 		/* index either some number, or "flexible" */
-		check_uniq("meta-disk", "%s:%s[%s]", host->name,
-			   host->meta_disk, host->meta_index);
+		for_each_host(h, host->on_hosts)
+			check_uniq("meta-disk", "%s:%s[%s]", h->name, host->meta_disk, host->meta_index);
 	} else if (host->meta_index) {
 		/* internal */
 		if (strcmp(host->meta_index, "flexible") != 0) {
@@ -196,44 +284,77 @@ void check_meta_disk(struct d_host_info *host)
 	}
 }
 
-#define EXP(TOKEN1)						\
-({ 								\
-	int token;						\
-	token = yylex();					\
-	if(token != TOKEN1)					\
-		pe_expected_got( #TOKEN1, token);		\
-	token;							\
-})
-
-#define EXP2(TOKEN1,TOKEN2)						\
-({ 									\
-	int token;							\
-	token = yylex();						\
-	if(token != TOKEN1 && token != TOKEN2)				\
-		pe_expected_got( #TOKEN1 "|" # TOKEN2, token);		\
-	token;								\
-})
-
 static void pe_expected(const char *exp)
 {
+	const char *s = yytext;
 	fprintf(stderr, "%s:%u: Parse error: '%s' expected,\n\t"
-		"but got '%s'\n", config_file, line, exp, yytext);
+		"but got '%.20s%s'\n", config_file, line, exp, s,
+		strlen(s) > 20 ? "..." : "");
+	exit(E_config_invalid);
+}
+
+static void check_string_error(int got)
+{
+	const char *msg;
+	switch(got) {
+	case TK_ERR_STRING_TOO_LONG:
+		msg = "Token too long";
+		break;
+	case TK_ERR_DQSTRING_TOO_LONG:
+		msg = "Double quoted string too long";
+		break;
+	case TK_ERR_DQSTRING:
+		msg = "Unterminated double quoted string\n  we don't allow embedded newlines\n ";
+		break;
+	default:
+		return;
+	}
+	fprintf(stderr,"%s:%u: %s >>>%.20s...<<<\n", config_file, line, msg, yytext);
 	exit(E_config_invalid);
 }
 
 static void pe_expected_got(const char *exp, int got)
 {
 	static char tmp[2] = "\0";
+	const char *s = yytext;
 	if (exp[0] == '\'' && exp[1] && exp[2] == '\'' && exp[3] == 0) {
 		tmp[0] = exp[1];
 	}
 	fprintf(stderr, "%s:%u: Parse error: '%s' expected,\n\t"
-		"but got '%s' (TK %d)\n",
+		"but got '%.20s%s' (TK %d)\n",
 		config_file, line,
-		tmp[0] ? tmp : exp, yytext, got);
+		tmp[0] ? tmp : exp, s, strlen(s) > 20 ? "..." : "", got);
 	exit(E_config_invalid);
 }
 
+#define EXP(TOKEN1)						\
+({								\
+	int token;						\
+	token = yylex();					\
+	if (token != TOKEN1) {					\
+		if (TOKEN1 == TK_STRING)			\
+			check_string_error(token);		\
+		pe_expected_got( #TOKEN1, token);		\
+	}							\
+	token;							\
+})
+
+static void expect_STRING_or_INT(void)
+{
+	int token = yylex();
+	switch(token) {
+	case TK_INTEGER:
+	case TK_STRING:
+		break;
+	case TK_ON:
+		yylval.txt = strdup(yytext);
+		break;
+	default:
+		check_string_error(token);
+		pe_expected_got("TK_STRING | TK_INTEGER", token);
+	}
+}
+
 static void parse_global(void)
 {
 	fline = line;
@@ -286,7 +407,20 @@ static void parse_global(void)
 	}
 }
 
-static struct d_option *parse_options(int token_switch, int token_option)
+static void check_and_change_deprecated_alias(char **name, int token_option)
+{
+	if (token_option == TK_HANDLER_OPTION) {
+		if (!strcmp(*name, "outdate-peer")) {
+			/* fprintf(stder, "config file:line: name is deprecated ...\n") */
+			free(*name);
+			*name = strdup("fence-peer");
+		}
+	}
+}
+
+static struct d_option *parse_options_d(int token_switch, int token_option,
+					int token_delegate, void (*delegate)(void*),
+					void *ctx)
 {
 	char *opt_name;
 	int token;
@@ -301,11 +435,15 @@ static struct d_option *parse_options(int token_switch, int token_option)
 			options = APPEND(options, new_opt(yylval.txt, NULL));
 		} else if (token == token_option) {
 			opt_name = yylval.txt;
+			check_and_change_deprecated_alias(&opt_name, token_option);
 			rc = yylval.rc;
-			EXP2(TK_STRING, TK_INTEGER);
+			expect_STRING_or_INT();
 			range_check(rc, opt_name, yylval.txt);
 			ro = new_opt(opt_name, yylval.txt);
 			options = APPEND(options, ro);
+		} else if (token == token_delegate) {
+			delegate(ctx);
+			continue;
 		} else if (token == '}') {
 			return options;
 		} else {
@@ -324,6 +462,82 @@ static struct d_option *parse_options(int token_switch, int token_option)
 	}
 }
 
+static struct d_option *parse_options(int token_switch, int token_option)
+{
+	return parse_options_d(token_switch, token_option, 0, NULL, NULL);
+}
+
+static void parse_address(struct d_name *on_hosts, char** addr, char** port, char** af)
+{
+	struct d_name *h;
+	switch(yylex()) {
+	case TK_SCI:
+	case TK_IPV4:
+		*af = yylval.txt;
+		EXP(TK_IPADDR);
+		break;
+	case TK_IPV6:
+		*af = yylval.txt;
+		EXP('[');
+		EXP(TK_IPADDR6);
+		break;
+	case TK_IPADDR:
+		*af = strdup("ipv4");
+		break;
+	/* case '[': // Do not foster people's laziness ;)
+		EXP(TK_IPADDR6);
+		*af = strdup("ipv6");
+		break; */
+	default:
+		pe_expected("sci | ipv4 | ipv6 | <ipv4 address> ");
+	}
+
+	*addr = yylval.txt;
+	if (!strcmp(*af, "ipv6"))
+		EXP(']');
+	EXP(':');
+	EXP(TK_INTEGER);
+	*port = yylval.txt;
+	range_check(R_PORT, "port", yylval.txt);
+	if (!strcmp(*addr, "127.0.0.1") || !strcmp(*addr, "::1"))
+		for_each_host(h, on_hosts)
+			check_uniq("IP", "%s:%s:%s", h->name, *addr, *port);
+	else
+		check_uniq("IP", "%s:%s", *addr, *port);
+	EXP(';');
+}
+
+static void parse_hosts(struct d_name **pnp, char delimeter)
+{
+	char errstr[20];
+	struct d_name *name;
+	int hosts = 0;
+	int token;
+
+	while (1) {
+		token = yylex();
+		switch (token) {
+		case TK_STRING:
+			name = malloc(sizeof(struct d_name));
+			name->name = yylval.txt;
+			name->next = NULL;
+			*pnp = name;
+			pnp = &name->next;
+			hosts++;
+			break;
+		default:
+			if (token == delimeter) {
+				if (!hosts)
+					pe_expected_got("TK_STRING", token);
+				return;
+			} else {
+				sprintf(errstr, "TK_STRING | '%c'", delimeter);
+				pe_expected_got(errstr, token);
+			}
+		}
+	}
+}
+
 static void parse_proxy_section(struct d_host_info *host)
 {
 	struct d_proxy_info *proxy;
@@ -332,27 +546,14 @@ static void parse_proxy_section(struct d_host_info *host)
 	host->proxy = proxy;
 
 	EXP(TK_ON);
-	EXP(TK_STRING);
-	proxy->name = yylval.txt;
-
-	EXP('{');
+	parse_hosts(&proxy->on_hosts, '{');
 	while (1) {
 		switch (yylex()) {
 		case TK_INSIDE:
-			EXP(TK_IPADDR);
-			proxy->inside_addr = yylval.txt;
-			EXP(':');
-			EXP(TK_INTEGER);
-			proxy->inside_port = yylval.txt;
-			EXP(';');
+			parse_address(proxy->on_hosts, &proxy->inside_addr, &proxy->inside_port, &proxy->inside_af);
 			break;
 		case TK_OUTSIDE:
-			EXP(TK_IPADDR);
-			proxy->outside_addr = yylval.txt;
-			EXP(':');
-			EXP(TK_INTEGER);
-			proxy->outside_port = yylval.txt;
-			EXP(';');
+			parse_address(proxy->on_hosts, &proxy->outside_addr, &proxy->outside_port, &proxy->outside_af);
 			break;
 		case '}':
 			goto break_loop;
@@ -361,8 +562,14 @@ static void parse_proxy_section(struct d_host_info *host)
 
 		}
 	}
-	
+
  break_loop:
+	if (!proxy->inside_addr)
+		pperror(host, proxy, "inside");
+
+	if (!proxy->outside_addr)
+		pperror(host, proxy, "outside");
+
 	return;
 }
 
@@ -381,38 +588,40 @@ static void parse_meta_disk(char **disk, char** index)
 	}
 }
 
-static void parse_host_section(struct d_resource *res, 
-			       char *host_name, int require_all)
+static void parse_host_section(struct d_resource *res,
+			       struct d_name* on_hosts, int require_all)
 {
 	struct d_host_info *host;
+	struct d_name *h;
 
 	c_section_start = line;
 	fline = line;
 
 	host=calloc(1,sizeof(struct d_host_info));
-	host->name = host_name;
-	check_uniq("host section", "%s: on %s", res->name, host->name);
+	host->on_hosts = on_hosts;
+	host->config_line = c_section_start;
+	for_each_host(h, on_hosts)
+		check_uniq("host section", "%s: on %s", res->name, h->name);
 	res->all_hosts = APPEND(res->all_hosts, host);
 
-	EXP('{');
 	while (1) {
 		switch (yylex()) {
 		case TK_DISK:
-			check_uniq("disk statement", "%s:%s:disk", res->name,
-				   host->name);
+			for_each_host(h, on_hosts)
+				check_uniq("disk statement", "%s:%s:disk", res->name, h->name);
 			EXP(TK_STRING);
 			host->disk = yylval.txt;
-			check_uniq("disk", "%s:%s:%s", "disk",
-				   host->name, yylval.txt);
+			for_each_host(h, on_hosts)
+				check_uniq("disk", "disk:%s:%s", h->name, yylval.txt);
 			EXP(';');
 			break;
 		case TK_DEVICE:
-			check_uniq("device statement", "%s:%s:device",
-				   res->name, host->name);
+			for_each_host(h, on_hosts)
+				check_uniq("device statement", "%s:%s:device", res->name, h->name);
 			EXP(TK_STRING);
 			host->device = yylval.txt;
-			check_uniq("device", "%s:%s:%s", "device",
-				   host->name, yylval.txt);
+			for_each_host(h, on_hosts)
+				check_uniq("device", "device:%s:%s", h->name, yylval.txt);
 			if (dt_minor_of_dev(host->device) < 0) {
 				fprintf(stderr, "%s:%d: cannot determine minor number of %s\n",
 					config_file, line, host->device);
@@ -421,26 +630,20 @@ static void parse_host_section(struct d_resource *res,
 			EXP(';');
 			break;
 		case TK_ADDRESS:
-			check_uniq("address statement", "%s:%s:address",
-				   res->name, host->name);
-			EXP(TK_IPADDR);
-			host->address = yylval.txt;
-			EXP(':');
-			EXP(TK_INTEGER);
-			host->port = yylval.txt;
-			range_check(R_PORT, "port", yylval.txt);
-			check_uniq("IP", "%s:%s", host->address, host->port);
-			EXP(';');
+			for_each_host(h, on_hosts)
+				check_uniq("address statement", "%s:%s:address", res->name, h->name);
+			parse_address(on_hosts, &host->address, &host->port, &host->address_family);
+			range_check(R_PORT, "port", host->port);
 			break;
 		case TK_META_DISK:
-			check_uniq("meta-disk statement", "%s:%s:meta-disk",
-				   res->name, host->name);
+			for_each_host(h, on_hosts)
+				check_uniq("meta-disk statement", "%s:%s:meta-disk", res->name, h->name);
 			parse_meta_disk(&host->meta_disk, &host->meta_index);
 			check_meta_disk(host);
 			break;
 		case TK_FLEX_META_DISK:
-			check_uniq("meta-disk statement", "%s:%s:meta-disk",
-				   res->name, host->name);
+			for_each_host(h, on_hosts)
+				check_uniq("meta-disk statement", "%s:%s:meta-disk", res->name, h->name);
 			EXP(TK_STRING);
 			host->meta_disk = yylval.txt;
 			if (strcmp("internal", yylval.txt)) {
@@ -471,17 +674,17 @@ static void parse_host_section(struct d_resource *res,
 	}
       break_loop:
 
-	/* Inerit device, disk, meta_disk and meta_index from the resource. */
+	/* Inherit device, disk, meta_disk and meta_index from the resource. */
 	if(!host->disk && res->disk) {
 		host->disk = strdup(res->disk);
-		check_uniq("disk", "%s:%s:%s", "disk",
-			   host->name, host->disk);
+		for_each_host(h, on_hosts)
+			check_uniq("disk", "disk:%s:%s", h->name, host->disk);
 	}
 
 	if(!host->device && res->device) {
 		host->device = strdup(res->device);
-		check_uniq("device", "%s:%s:%s", "device",
-			   host->name, host->device);
+		for_each_host(h, on_hosts)
+			check_uniq("device", "device:%s:%s", h->name, host->device);
 	}
 
 	if(!host->meta_disk && res->meta_disk) {
@@ -505,15 +708,18 @@ static void parse_host_section(struct d_resource *res,
 void parse_skip()
 {
 	int level;
+	int token;
 	fline = line;
 
-	switch (yylex()) {
+	token = yylex();
+	switch (token) {
 	case TK_STRING:
 		EXP('{');
 		break;
 	case '{':
 		break;
 	default:
+		check_string_error(token);
 		pe_expected("[ some_text ] {");
 	}
 
@@ -538,17 +744,118 @@ void parse_skip()
 	while (level) ;
 }
 
+static void parse_stacked_section(struct d_resource* res)
+{
+	struct d_host_info *host;
+	struct d_resource *l_res, *tmp;
+	struct d_name *h;
+	char *l_res_name;
+
+	c_section_start = line;
+	fline = line;
+
+	host=calloc(1,sizeof(struct d_host_info));
+	res->all_hosts = APPEND(res->all_hosts, host);
+	EXP(TK_STRING);
+	l_res_name = yylval.txt;
+	check_uniq("stacked-on-top-of", "stacked:%s", l_res_name);
+
+	for_each_resource(l_res, tmp, config) {
+		if (!strcmp(l_res->name, l_res_name))
+			break;
+	}
+	if (l_res == NULL) {
+		fprintf(stderr, "%s:%d: in resource %s, "
+			"referenced resource '%s' not yet defined.\n",
+			config_file, c_section_start, res->name, l_res_name);
+		exit(E_config_invalid);
+	}
+	if (l_res->stacked) {
+		fprintf(stderr,
+			"%s:%d: in resource %s, stacked-on-top-of %s { ... }:\n"
+			"\tFIXME. I won't stack stacked resources.\n",
+			config_file, c_section_start, res->name, l_res_name);
+		exit(E_config_invalid);
+	}
+
+	host->on_hosts = concat_names(l_res->me->on_hosts, l_res->peer->on_hosts);
+	host->lower = l_res;
+
+	m_asprintf(&host->meta_disk, "%s", "internal");
+	m_asprintf(&host->meta_index, "%s", "internal");
+	m_asprintf(&host->disk, "%s", l_res->me->device);
+
+	EXP('{');
+	while (1) {
+		switch(yylex()) {
+		case TK_DEVICE:
+			EXP(TK_STRING);
+			host->device = yylval.txt;
+			for_each_host(h, host->on_hosts)
+				check_uniq("device", "device:%s:%s", h->name, yylval.txt);
+			EXP(';');
+			break;
+		case TK_ADDRESS:
+			for_each_host(h, host->on_hosts)
+				check_uniq("address statement", "%s:%s:address", res->name, h->name);
+			parse_address(l_res->me->on_hosts, &host->address, &host->port, &host->address_family);
+			range_check(R_PORT, "port", yylval.txt);
+			break;
+		case TK_PROXY:
+			parse_proxy_section(host);
+			break;
+		case '}':
+			goto break_loop;
+		default:
+			pe_expected("device | address | proxy");
+		}
+	}
+ break_loop:
+
+	/* inherit device */
+	if (!host->device && res->device) {
+		host->device = strdup(res->device);
+		for_each_host(h, host->on_hosts)
+			check_uniq("device", "device:%s:%s", h->name, host->device);
+	}
+
+	if (!host->device)
+		derror(host,res,"device");
+	if (!host->disk)
+		derror(host,res,"disk");
+	if (!host->address)
+		derror(host,res,"address");
+	if (!host->meta_disk)
+		derror(host,res,"meta-disk");
+}
+
+void startup_delegate(void *ctx)
+{
+	struct d_resource *res = (struct d_resource *)ctx;
+
+	if (!strcmp(yytext, "become-primary-on")) {
+		parse_hosts(&res->become_primary_on, ';');
+	} else if (!strcmp(yytext, "stacked-timeouts")) {
+		res->stacked_timeouts = 1;
+		EXP(';');
+	} else
+		pe_expected("<an option keyword> | become-primary-on | stacked-timeouts");
+}
+
 struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 {
 	struct d_resource* res;
 	struct d_host_info *host;
+	struct d_name *host_names;
 	int token;
 
 	fline = line;
 
+	check_uniq("resource section", res_name);
+
 	res=calloc(1,sizeof(struct d_resource));
 	res->name = res_name;
-	res->next = NULL;
+	res->me_minor = -1; /* will be set once in dt_minor_of_res */
 
 	while(1) {
 		switch((token=yylex())) {
@@ -559,14 +866,35 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 			EXP(';');
 			break;
 		case TK_ON:
+			parse_hosts(&host_names, '{');
+			parse_host_section(res, host_names, 1);
+			break;
+		case TK_STACKED:
+			parse_stacked_section(res);
+			break;
+		case TK_IGNORE:
+			if (res->me || res->peer) {
+				fprintf(stderr,
+					"%s:%d: in resource %s, "
+					"'ignore-on' statement must precede any real host section (on ... { ... }).\n",
+					config_file, line, res->name);
+				exit(E_config_invalid);
+			}
 			EXP(TK_STRING);
-			parse_host_section(res, yylval.txt, 1);
+			fprintf(stderr, "%s:%d: in resource %s, "
+			       "WARN: The 'ignore-on' keyword is deprecated.\n",
+			       config_file, line, res->name);
+			EXP(';');
 			break;
 		case TK__THIS_HOST:
-			parse_host_section(res, strdup("_this_host"), 0);
+			EXP('{');
+			host_names = names_from_str("_this_host");
+			parse_host_section(res, host_names, 0);
 			break;
 		case TK__REMOTE_HOST:
-			parse_host_section(res, strdup("_remote_host"), 0);
+			EXP('{');
+			host_names = names_from_str("_remote_host");
+			parse_host_section(res, host_names, 0);
 			break;
 		case TK_DISK:
 			switch (token=yylex()) {
@@ -579,7 +907,8 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 				res->disk_options = parse_options(TK_DISK_SWITCH,
 								  TK_DISK_OPTION);
 				break;
-			default: 
+			default:
+				check_string_error(token);
 				pe_expected_got( "TK_STRING | {", token);
 			}
 			break;
@@ -598,8 +927,11 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 		case TK_STARTUP:
 			check_uniq("startup section", "%s:startup", res->name);
 			EXP('{');
-			res->startup_options=parse_options(TK_STARTUP_SWITCH,
-							   TK_STARTUP_OPTION);
+			res->startup_options=parse_options_d(TK_STARTUP_SWITCH,
+							     TK_STARTUP_OPTION,
+							     TK_STARTUP_DELEGATE,
+							     &startup_delegate,
+							     res);
 			break;
 		case TK_HANDLER:
 			check_uniq("handlers section", "%s:handlers", res->name);
@@ -633,7 +965,8 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 			goto exit_loop;
 		default:
 			pe_expected_got("protocol | on | disk | net | syncer |"
-					" startup | handlers",token);
+					" startup | handlers |"
+					" ignore-on | stacked-on-top-of",token);
 		}
 	}
 
@@ -641,32 +974,59 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 
 	/* Determin the local host section and the peer host section. */
 	host = res->all_hosts;
-	while(host) {
-		if(!strcmp(host->name, nodeinfo.nodename) ||
-		   !strcmp(host->name, "_this_host") ||
-		   ( host->proxy && 
-		     !strcmp(host->proxy->name, nodeinfo.nodename))) {
-			if(res->me) {
+	while (host) {
+		if (!res->ignore && (res->me && res->peer)) {
+			fprintf(stderr,
+				"%s:%d: in resource %s, "
+				"unsupported third host section %s %s { ... }.\n",
+				config_file, host->config_line, res->name,
+				host->lower ? "stacked-on-top-of" : "on",
+				host->lower ? host->lower->name : names_to_str(host->on_hosts));
+			exit(E_config_invalid);
+		}
+
+		if (name_in_names(nodeinfo.nodename, host->on_hosts) ||
+		    name_in_names("_this_host", host->on_hosts) ||
+		    ( host->proxy && name_in_names(nodeinfo.nodename, host->proxy->on_hosts))) {
+			if (res->ignore) {
 				config_valid = 0;
 				fprintf(stderr,
-					"%s:%d: in resource %s, on %s { ... } ... on %s { ... }:\n"
-					"\tThere are multiple host sections for this node.\n"
-					"\tMaybe misspelled local host name '%s'?\n",
-					config_file, c_section_start, res->name,
-						res->me->name, host->name, nodeinfo.nodename);
+					"%s:%d: in resource %s, %s %s { ... }:\n"
+					"\tYou cannot ignore and define at the same time.\n",
+					config_file, host->config_line, res->name,
+					host->lower ? "stacked-on-top-of" : "on",
+					host->lower ? host->lower->name : names_to_str(host->on_hosts));
+			}
+			if (res->me) {
+				config_valid = 0;
+				fprintf(stderr,
+					"%s:%d: in resource %s, %s %s { ... } ... %s %s { ... }:\n"
+					"\tThere are multiple host sections for this node.\n",
+					config_file, host->config_line, res->name,
+					res->me->lower ? "stacked-on-top-of" : "on",
+					res->me->lower ? res->me->lower->name : names_to_str(res->me->on_hosts),
+					host->lower ? "stacked-on-top-of" : "on",
+					host->lower ? host->lower->name : names_to_str(host->on_hosts));
 			}
 			res->me = host;
+			if (host->lower)
+				res->stacked = 1;
 		} else {
-			/* This needs to be refined as soon as we support 
+			/* This needs to be refined as soon as we support
 			   multiple peers in the resource section */
-			if(res->peer) {
-				config_valid = 0;
-				fprintf(stderr,
-					"%s:%d: in resource %s, on %s { ... } ... on %s { ... }:\n"
-					"\tThere are multiple host sections for the peer.\n"
-					"\tMaybe misspelled local host name '%s'?\n",
-					config_file, c_section_start, res->name,
-						res->peer->name, host->name, nodeinfo.nodename);
+			if (res->peer) {
+				if (!res->me) {
+					/* just store it anyways into ->me */
+					/* reorder, so it dumps out in the same order as read in */
+					res->me = res->peer;
+					res->peer = host;
+					res->ignore = 1; /* implicit ignore */
+				} else {
+					/* hm. if that did not work, I cannot ignore it */
+					config_valid = 0;
+					fprintf(stderr, "THINKO 1\n");
+					exit(E_thinko);
+				}
 			}
 			res->peer = host;
 		}
@@ -721,7 +1081,7 @@ void my_parse(void)
 		case TK_RESOURCE:
 			EXP(TK_STRING);
 			EXP('{');
-			config=APPEND(config, 
+			config=APPEND(config,
 				      parse_resource(yylval.txt,BothHRequired));
 			break;
 		case TK_SKIP:
