@@ -50,6 +50,8 @@ enum yytokentype {
 	TK_GLOBAL = 258,
 	TK_RESOURCE,
 	TK_ON,
+	TK_STACKED,
+	TK_IGNORE,
 	TK_NET,
 	TK_DISK,
 	TK_SKIP,
@@ -58,19 +60,17 @@ enum yytokentype {
 	TK_DISABLE_IP_VERIFICATION,
 	TK_DIALOG_REFRESH,
 	TK_PROTOCOL,
-	TK_HANDLER, // 270
+	TK_HANDLER,
 	TK_COMMON,
 	TK_ADDRESS,
 	TK_DEVICE,
 	TK_META_DISK,
 	TK_FLEX_META_DISK,
 	TK_MINOR_COUNT,
-
 	TK_IPADDR,
 	TK_INTEGER,
 	TK_STRING,
-	TK_ELSE, // 280
-
+	TK_ELSE,
 	TK_DISK_SWITCH,
 	TK_DISK_OPTION,
 	TK_NET_SWITCH,
@@ -79,8 +79,9 @@ enum yytokentype {
 	TK_SYNCER_OPTION,
 	TK_STARTUP_SWITCH,
 	TK_STARTUP_OPTION,
+	TK_STARTUP_DELEGATE,
 	TK_HANDLER_OPTION,
-	TK_USAGE_COUNT,  // 290
+	TK_USAGE_COUNT,
 	TK_ASK,
 	TK_YES,
 	TK_NO,
@@ -95,6 +96,13 @@ enum yytokentype {
 	TK_MEMLIMIT,
 	TK_PROXY_OPTION,
 	TK_PROXY_SWITCH,
+	TK_ERR_STRING_TOO_LONG,
+	TK_ERR_DQSTRING_TOO_LONG,
+	TK_ERR_DQSTRING,
+	TK_SCI,
+	TK_IPV4,
+	TK_IPV6,
+	TK_IPADDR6,
 };
 
 typedef struct YYSTYPE {
