@@ -45,6 +45,7 @@
 #include <sys/time.h>
 #include <time.h>
 
+#define __bitwise /* Build-workaround for broken RHEL4 kernels (2.6.9_78.0.1) */
 #include <linux/netlink.h>
 #include <linux/connector.h>
 
@@ -141,6 +142,7 @@ int call_drbd(int sk_nl, struct drbd_tag_list *tl, struct nlmsghdr* nl_hdr,
 void close_cn(int sk_nl);
 
 // other functions
+int get_af_sci(int warn);
 void print_command_usage(int i, const char *addinfo, enum usage_type);
 
 // command functions
@@ -168,7 +170,9 @@ void string_opt_xml(struct drbd_option *option);
 
 // sub commands for generic_get_cmd
 int show_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
-int state_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
+int role_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
+int status_xml_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
+int sh_status_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
 int cstate_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
 int dstate_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
 int uuids_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl);
@@ -181,6 +185,7 @@ int conv_protocol(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
 
 // convert functions for options
 int conv_numeric(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
+int conv_sndbuf(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
 int conv_handler(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
 int conv_bit(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
 int conv_string(struct drbd_option *od, struct drbd_tag_list *tl, char* arg);
@@ -249,6 +254,10 @@ struct option wait_cmds_options[] = {
 	conv_numeric, show_numeric, numeric_opt_usage, numeric_opt_xml, \
 	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
 		DRBD_ ## N ## _DEF ,U,UN  } }
+#define EN_sndbuf(N,U,UN) \
+	conv_sndbuf, show_numeric, numeric_opt_usage, numeric_opt_xml, \
+	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
+		DRBD_ ## N ## _DEF ,U,UN  } }
 #define EH(N,D) \
 	conv_handler, show_handler, handler_opt_usage, handler_opt_xml, \
 	{ .handler_param = { N, ARRY_SIZE(N), \
@@ -280,7 +289,9 @@ struct drbd_cmd commands[] = {
 		 { "on-io-error",'e',	T_on_io_error,	EH(on_error,ON_IO_ERROR) },
 		 { "fencing",'f',	T_fencing,      EH(fencing_n,FENCING) },
 		 { "use-bmbv",'b',	T_use_bmbv,     EB },
+		 { "no-disk-barrier",'a',T_no_disk_barrier,EB },
 		 { "no-disk-flushes",'i',T_no_disk_flush,EB },
+		 { "no-disk-drain",'D', T_no_disk_drain,EB },
 		 { "no-md-flushes",'m', T_no_md_flush,  EB },
 		 { "max-bio-bvecs",'s',	T_max_bio_bvecs,EN(MAX_BIO_BVECS,1,NULL) },
 		 CLOSE_OPTIONS }} }, },
@@ -289,8 +300,8 @@ struct drbd_cmd commands[] = {
 
 	{"net", P_net_conf, F_CONFIG_CMD, {{
 	 (struct drbd_argument[]) {
-		 { "local_addr",	T_my_addr,	conv_address },
-		 { "remote_addr",	T_peer_addr,	conv_address },
+		 { "[af:]local_addr[:port]",T_my_addr,	conv_address },
+		 { "[af:]remote_addr[:port]",T_peer_addr,conv_address },
 		 { "protocol",		T_wire_protocol,conv_protocol },
 		 { NULL,                0,           	NULL}, },
 	 (struct drbd_option[]) {
@@ -300,7 +311,7 @@ struct drbd_cmd commands[] = {
 		 { "unplug-watermark",'u',T_unplug_watermark, EN(UNPLUG_WATERMARK,1,NULL) },
 		 { "connect-int",'c',	T_try_connect_int, EN(CONNECT_INT,1,"seconds") },
 		 { "ping-int",'i',	T_ping_int,	   EN(PING_INT,1,"seconds") },
-		 { "sndbuf-size",'S',	T_sndbuf_size,	   EN(SNDBUF_SIZE,1,"bytes") },
+		 { "sndbuf-size",'S',	T_sndbuf_size,	   EN_sndbuf(SNDBUF_SIZE,1,"bytes") },
 		 { "ko-count",'k',	T_ko_count,	   EN(KO_COUNT,1,NULL) },
 		 { "allow-two-primaries",'m',T_two_primaries, EB },
 		 { "cram-hmac-alg",'a',	T_cram_hmac_alg,   ES },
@@ -313,6 +324,7 @@ struct drbd_cmd commands[] = {
 		 { "ping-timeout",'p',  T_ping_timeo,	   EN(PING_TIMEO,1,"1/10 seconds") },
 		 { "discard-my-data",'D', T_want_lose,     EB },
 		 { "data-integrity-alg",'d', T_integrity_alg,     ES },
+		 { "no-tcp-cork",'o',   T_no_cork,         EB },
 		 CLOSE_OPTIONS }} }, },
 
 	{"disconnect", P_disconnect, F_CONFIG_CMD, {{NULL, NULL}} },
@@ -327,10 +339,16 @@ struct drbd_cmd commands[] = {
 		 { "rate",'r',T_rate,			EN(RATE,'k',"bytes/second") },
 		 { "after",'a',T_after,			EN(AFTER,1,NULL) },
 		 { "al-extents",'e',T_al_extents,	EN(AL_EXTENTS,1,NULL) },
+		 { "csums-alg", 'C',T_csums_alg,        ES },
 		 { "verify-alg", 'v',T_verify_alg,      ES },
 		 { "cpu-mask",'c',T_cpu_mask,           ES },
 		 CLOSE_OPTIONS }} }, },
 
+	{"new-current-uuid", P_new_c_uuid, F_CONFIG_CMD, {{NULL,
+	 (struct drbd_option[]) {
+		 { "clear-bitmap",'c',T_clear_bm, EB   },
+		 CLOSE_OPTIONS }} }, },
+
 	{"invalidate", P_invalidate, F_CONFIG_CMD, {{ NULL, NULL }} },
 	{"invalidate-remote", P_invalidate_peer, F_CONFIG_CMD, {{NULL, NULL}} },
 	{"pause-sync", P_pause_sync, F_CONFIG_CMD, {{ NULL, NULL }} },
@@ -340,7 +358,11 @@ struct drbd_cmd commands[] = {
 	{"outdate", P_outdate, F_CONFIG_CMD, {{ NULL, NULL }} },
 	{"verify", P_start_ov, F_CONFIG_CMD, {{NULL, NULL}} },
 	{"down",            0, down_cmd, get_usage, { {NULL, NULL }} },
-	{"state", P_get_state, F_GET_CMD, { .gp={ state_scmd} } },
+	/* "state" is deprecated! please use "role".
+	 * find_cmd_by_name still understands "state", however. */
+	{"role", P_get_state, F_GET_CMD, { .gp={ role_scmd} } },
+	{"status", P_get_state, F_GET_CMD, {.gp={ status_xml_scmd } } },
+	{"sh-status", P_get_state, F_GET_CMD, {.gp={ sh_status_scmd } } },
 	{"cstate", P_get_state, F_GET_CMD, {.gp={ cstate_scmd} } },
 	{"dstate", P_get_state, F_GET_CMD, {.gp={ dstate_scmd} } },
 	{"show-gi", P_get_uuids, F_GET_CMD, {.gp={ uuids_scmd} }},
@@ -414,6 +436,10 @@ static const char *error_messages[] = {
 	EM(VERIFYAlgNotDigest) = "VERIFYAlgNotDigest",
 	EM(VERIFYIsRunning) = "Can not change verify-alg while online verify runs",
 	EM(DataOfWrongCurrent) = "Can only attach to the data we lost last (see kernel log).",
+	EM(MayNotBeConnected) = "Not possible while connected",
+	EM(CSUMSAlgNotAvail) = "CSUMSAlgNotAvail",
+	EM(CSUMSAlgNotDigest) = "CSUMSAlgNotDigest",
+	EM(CSUMSResyncRunning) = "Can not change csums-alg while resync is in progress",
 };
 #define MAX_ERROR (sizeof(error_messages)/sizeof(*error_messages))
 const char * error_to_string(int err_no)
@@ -426,8 +452,11 @@ const char * error_to_string(int err_no)
 
 char *cmdname = NULL; /* "drbdsetup" for reporting in usage etc. */
 char *devname = NULL; /* "/dev/drbd12" for reporting in print_config_error */
+char *resname = NULL; /* for pretty printing in "status" only,
+			 taken from environment variable DRBD_RESOURCE */
 int debug_dump_argv = 0; /* enabled by setting DRBD_DEBUG_DUMP_ARGV in the environment */
 int lock_fd;
+unsigned int cn_idx;
 
 int dump_tag_list(unsigned short *tlc)
 {
@@ -589,6 +618,24 @@ int port_part(const char* s)
 	return 7788;
 }
 
+void resolv6(char *name, struct in6_addr *addr)
+{
+	int rv;
+	struct hostent *he;
+
+	rv = inet_pton(AF_INET6, name, addr);
+	if (rv > 0)
+		return;
+	else if (rv == 0) {
+		he = gethostbyname2(name, AF_INET6);
+		if (!he) {
+			PERROR("can not resolv the hostname");
+			exit(20);
+		}
+		memcpy(addr, he->h_addr_list[0], sizeof(struct in6_addr));
+	}
+}
+
 unsigned long resolv(const char* name)
 {
 	unsigned long retval;
@@ -605,16 +652,70 @@ unsigned long resolv(const char* name)
 	return retval;
 }
 
+static void split_address(char* text, int *af, char** address, int* port)
+{
+	static struct { char* text; int af; } afs[] = {
+		{ "ipv4:", AF_INET  },
+		{ "ipv6:", AF_INET6 },
+		{ "sci:",  -1 },
+	};
+
+	unsigned int i;
+	char *b;
+
+	*af=AF_INET;
+	*address = text;
+	for (i=0; i<ARRY_SIZE(afs); i++) {
+		if (!strncmp(text, afs[i].text, strlen(afs[i].text))) {
+			*af = afs[i].af;
+			*address = text + strlen(afs[i].text);
+			break;
+		}
+	}
+	if (*af == -1)
+		*af = get_af_sci(1);
+
+	b=strrchr(text,':');
+	if (b) {
+		*b = 0;
+		*port = m_strtoll(b+1,1);
+	} else
+		*port = 7788;
+}
+
 int conv_address(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg)
 {
+	static int mind_af_set = 0;
 	struct sockaddr_in addr;
-	int addr_len = sizeof(struct sockaddr_in);
+	struct sockaddr_in6 addr6;
+	int af, port;
+	char *address, bit=0;
 
-	addr.sin_port = htons(port_part(arg));
-	addr.sin_family = AF_INET;
-	addr.sin_addr.s_addr = resolv(addr_part(arg));
+	split_address(arg, &af, &address, &port);
 
-	add_tag(tl,ad->tag,&addr,addr_len);
+	/* The mind_af tag is mandatory. I.e. the module may not silently ignore it.
+	   That means that an older DRBD module must fail the operation since it does
+	   not know the mind_af tag. We set it in case we use an other AF then AF_INET,
+	   so that the alternate AF is not silently ignored by the DRBD module */
+	if (af != AF_INET && !mind_af_set) {
+		add_tag(tl,T_mind_af,&bit,sizeof(bit));
+		mind_af_set=1;
+	}
+
+	if (af == AF_INET6) {
+		memset(&addr6, 0, sizeof(struct sockaddr_in6));
+		addr6.sin6_port = htons(port);
+		addr6.sin6_family = AF_INET6;
+		resolv6(address, &addr6.sin6_addr);
+		/* addr6.sin6_len = sizeof(addr6); */
+		add_tag(tl,ad->tag,&addr6,sizeof(addr6));
+	} else {
+		/* AF_INET and AF_SCI */
+		addr.sin_port = htons(port);
+		addr.sin_family = af;
+		addr.sin_addr.s_addr = resolv(address);
+		add_tag(tl,ad->tag,&addr,sizeof(addr));
+	}
 
 	return NoError;
 }
@@ -648,6 +749,59 @@ int conv_bit(struct drbd_option *od, struct drbd_tag_list *tl, char* arg __attri
 	return NoError;
 }
 
+/* It will only print the WARNING if the warn flag is set
+   with the _first_ call! */
+#define PROC_NET_AF_SCI_FAMILY "/proc/net/af_sci/family"
+int get_af_sci(int warn)
+{
+	char buf[16];
+	int c, fd;
+	static int af = -1;
+
+	if (af > 0)
+		return af;
+
+	fd = open(PROC_NET_AF_SCI_FAMILY, O_RDONLY);
+	if (fd < 0) {
+		if (warn)
+			fprintf(stderr, "open(" PROC_NET_AF_SCI_FAMILY ") "
+				"failed: %m\n WARNING: assuming AF_SCI = 27. "
+				"Socket creation will probabely fail.\n");
+		af = 27;
+		return af;
+	}
+	c = read(fd, buf, sizeof(buf)-1);
+	if (c > 0) {
+		buf[c] = 0;
+		if (buf[c-1] == '\n')
+			buf[c-1] = 0;
+		af = m_strtoll(buf,1);
+	} else {
+		if (warn)
+			fprintf(stderr, "read(" PROC_NET_AF_SCI_FAMILY ") "
+				"failed: %m\n WARNING: assuming AF_SCI = 27. "
+				"Socket creation will probabely fail.\n");
+		af = 27;
+	}
+	close(fd);
+	return af;
+}
+
+int conv_sndbuf(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
+{
+	int err = conv_numeric(od, tl, arg);
+	long long l = m_strtoll(arg, 0);
+	char bit = 0;
+
+	if (err != NoError || l != 0)
+		return err;
+	/* this is a mandatory bit,
+	 * to avoid newer userland to configure older modules with
+	 * a sndbuf size of zero, which would lead to Oops. */
+	add_tag(tl, T_auto_sndbuf_size, &bit, sizeof(bit));
+	return NoError;
+}
+
 int conv_numeric(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
 {
 	const long long min = od->numeric_param.min;
@@ -678,7 +832,6 @@ int conv_numeric(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
 		fprintf(stderr, "internal error in conv_numeric()\n");
 	}
 	return NoError;
-
 }
 
 int conv_handler(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
@@ -711,11 +864,12 @@ int conv_string(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
 	return NoError;
 }
 
+
 struct option *	make_longoptions(struct drbd_option* od)
 {
 	/* room for up to N options,
 	 * plus set-defaults, create-device, and the terminating NULL */
-	const int N = 20;
+#define N 20
 	static struct option buffer[N+3];
 	int i=0;
 
@@ -732,6 +886,7 @@ struct option *	make_longoptions(struct drbd_option* od)
 		}
 		od++;
 	}
+#undef N
 
 	// The two omnipresent options:
 	buffer[i].name = "set-defaults";
@@ -836,18 +991,18 @@ void dump_argv(int argc, char **argv, int first_non_option, int n_known_args)
 	int i;
 	if (!debug_dump_argv)
 		return;
-	printf(",-- ARGV dump (optind %d, known_args %d, argc %u):\n",
+	fprintf(stderr, ",-- ARGV dump (optind %d, known_args %d, argc %u):\n",
 		first_non_option, n_known_args, argc);
 	for (i = 0; i < argc; i++) {
 		if (i == 1)
-			puts("-- consumed options:");
+			fprintf(stderr, "-- consumed options:");
 		if (i == first_non_option)
-			puts("-- known args:");
+			fprintf(stderr, "-- known args:");
 		if (i == (first_non_option + n_known_args))
-			puts("-- unexpected args:");
-		printf("| %2u: %s\n", i, argv[i]);
+			fprintf(stderr, "-- unexpected args:");
+		fprintf(stderr, "| %2u: %s\n", i, argv[i]);
 	}
-	printf("`--\n");
+	fprintf(stderr, "`--\n");
 }
 
 int _generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
@@ -871,7 +1026,7 @@ int _generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 			rv = OTHER_ERROR;
 			goto error;
 		}
-		rv |= ad->convert_function(ad,tl,argv[i++]);
+		rv = ad->convert_function(ad,tl,argv[i++]);
 		if (rv != NoError)
 			goto error;
 		ad++;
@@ -882,7 +1037,8 @@ int _generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 	opterr=0;
 	while( (c=getopt_long(argc,argv,make_optstring(lo,':'),lo,0)) != -1 ) {
 		od = find_opt_by_short_name(cm->cp.options,c);
-		if(od) rv |= od->convert_function(od,tl,optarg);
+		if (od)
+			rv = od->convert_function(od,tl,optarg);
 		else {
 			if(c=='(') flags |= DRBD_NL_SET_DEFAULTS;
 			else if(c==')') flags |= DRBD_NL_CREATE_DEVICE;
@@ -947,6 +1103,27 @@ int generic_config_cmd(struct drbd_cmd *cm, int minor, int argc, char **argv)
 #define ASSERT(exp) if (!(exp)) \
 		fprintf(stderr,"ASSERT( " #exp " ) in %s:%d\n", __FILE__,__LINE__);
 
+void show_af(struct drbd_option *od, unsigned short* tp)
+{
+	int af_sci = get_af_sci(0);
+	int val;
+	const char *msg;
+
+	ASSERT(tag_type(*tp++) == TT_INTEGER);
+	ASSERT( *tp++ == sizeof(int) );
+	val = *(int*)tp;
+
+	msg = (val == af_sci) ? "sci" :
+	      (val == AF_INET) ? "IPv4" :
+	      "UNKNOWN";
+	printf("\t%-16s\t%s", od->name, msg);
+	if (val == AF_INET) printf(" _is_default");
+	if (val != AF_INET && val != af_sci)
+		printf("; # %u ??\n", val);
+	else
+		printf(";\n");
+}
+
 void show_numeric(struct drbd_option *od, unsigned short* tp)
 {
 	long long val;
@@ -999,7 +1176,7 @@ void show_bit(struct drbd_option *od, unsigned short* tp)
 void show_string(struct drbd_option *od, unsigned short* tp)
 {
 	ASSERT( tag_type(*tp++) == TT_STRING );
-	if( *tp++ > 0) printf("\t%-16s\t\"%s\";\n",od->name,(char*)tp);
+	if( *tp++ > 0 && *(char*)tp) printf("\t%-16s\t\"%s\";\n",od->name,(char*)tp);
 }
 
 unsigned short *look_for_tag(unsigned short *tlc, unsigned short tag)
@@ -1127,6 +1304,7 @@ int generic_get_cmd(struct drbd_cmd *cm, int minor, int argc,
 	struct drbd_tag_list *tl;
 	struct drbd_nl_cfg_reply *reply;
 	int sk_nl,rv;
+	int ignore_minor_not_known;
 
 	if (argc > 1)
 		warn_print_excess_args(argc, argv, 1);
@@ -1143,13 +1321,21 @@ int generic_get_cmd(struct drbd_cmd *cm, int minor, int argc,
 	tl->drbd_p_header->drbd_minor = minor;
 	tl->drbd_p_header->flags = 0;
 
+	memset(buffer,0,sizeof(buffer));
 	call_drbd(sk_nl,tl, (struct nlmsghdr*)buffer,4096,NL_TIME);
 
 	close_cn(sk_nl);
 	reply = (struct drbd_nl_cfg_reply *)
 		((struct cn_msg *)NLMSG_DATA(buffer))->data;
 
-	if (reply->ret_code != NoError)
+	/* if there was an error, report and abort --
+	 * unless it was "this device is not there",
+	 * and command was "status" */
+	ignore_minor_not_known =
+		cm->gp.show_function == status_xml_scmd ||
+		cm->gp.show_function == sh_status_scmd;
+	if (reply->ret_code != NoError &&
+	   !(reply->ret_code == MinorNotKnown && ignore_minor_not_known))
 		return print_config_error(reply->ret_code);
 
 	rv = cm->gp.show_function(cm,minor,reply->tag_list);
@@ -1162,11 +1348,47 @@ int generic_get_cmd(struct drbd_cmd *cm, int minor, int argc,
 	return rv;
 }
 
+char *af_to_str(int af)
+{
+	if (af == AF_INET)
+		return "ipv4";
+	else if (af == AF_INET6)
+		return "ipv6";
+	else if (af == get_af_sci(0))
+		return "sci";
+	else return "unknown";
+}
+
+void show_address(void* address, int addr_len)
+{
+	struct sockaddr     *addr;
+	struct sockaddr_in  *addr4;
+	struct sockaddr_in6 *addr6;
+	char buffer[INET6_ADDRSTRLEN];
+
+	addr = (struct sockaddr *)address;
+	if (addr->sa_family == AF_INET || addr->sa_family == get_af_sci(0)) {
+		addr4 = (struct sockaddr_in *)address;
+		printf("\taddress\t\t\t%s %s:%d;\n",
+		       af_to_str(addr4->sin_family),
+		       inet_ntoa(addr4->sin_addr),
+		       ntohs(addr4->sin_port));
+	} else if (addr->sa_family == AF_INET6) {
+		addr6 = (struct sockaddr_in6 *)address;
+		printf("\taddress\t\t\t%s [%s]:%d;\n",
+		       af_to_str(addr6->sin6_family),
+		       inet_ntop(addr6->sin6_family, &addr6->sin6_addr, buffer, INET6_ADDRSTRLEN),
+		       ntohs(addr6->sin6_port));
+	} else {
+		printf("\taddress\t\t\t[unknown af=%d, len=%d]\n", addr->sa_family, addr_len);
+	}
+}
+
 int show_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl)
 {
-	int idx;
+	int idx = idx;
 	char *str, *backing_dev, *address;
-	struct sockaddr_in *addr;
+	unsigned int addr_len = addr_len;
 
 	// find all commands that have options and print those...
 	for ( cm = commands ; cm < commands + ARRY_SIZE(commands) ; cm++ ) {
@@ -1179,7 +1401,7 @@ int show_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl)
 		printf("protocol %c;\n",'A'+idx-1);
 	backing_dev = address = NULL;
 	consume_tag_string(T_backing_dev,rtl,&backing_dev);
-	consume_tag_string(T_my_addr,rtl,&address);
+	consume_tag_blob(T_my_addr, rtl, &address, &addr_len);
 	if(backing_dev || address) {
 		printf("_this_host {\n");
 		printf("\tdevice\t\t\t\"/dev/drbd%d\";\n",minor);
@@ -1198,34 +1420,150 @@ int show_scmd(struct drbd_cmd *cm, int minor, unsigned short *rtl)
 			default:
 				printf("\tmeta-disk\t\t\"%s\" [ %d ];\n",str,
 				       idx);
-			}
-		}
-		if(address) {
-			addr = (struct sockaddr_in *)address;
-			printf("\taddress\t\t\t%s:%d;\n",
-			       inet_ntoa(addr->sin_addr),
-			       ntohs(addr->sin_port));
+			 }
 		}
+		if(address)
+			show_address(address, addr_len);
 		printf("}\n");
 	}
 
-	if(consume_tag_string(T_peer_addr,rtl,&str)) {
+	if(consume_tag_blob(T_peer_addr, rtl, &address, &addr_len)) {
 		printf("_remote_host {\n");
-		addr = (struct sockaddr_in *)str;
-		printf("\taddress\t\t\t%s:%d;\n",
-		       inet_ntoa(addr->sin_addr),
-		       ntohs(addr->sin_port));
+		show_address(address, addr_len);
 		printf("}\n");
 	}
+	consume_tag_bit(T_mind_af, rtl, &idx); /* consume it, its value has no relevance */
+	consume_tag_bit(T_auto_sndbuf_size, rtl, &idx); /* consume it, its value has no relevance */
 
 	return 0;
 }
 
-int state_scmd(struct drbd_cmd *cm __attribute((unused)),
+int status_xml_scmd(struct drbd_cmd *cm __attribute((unused)),
+		int minor, unsigned short *rtl)
+{
+	union drbd_state_t state = { .i = 0 };
+	int synced = 0;
+
+	if (!consume_tag_int(T_state_i,rtl,(int*)&state.i)) {
+		printf( "<!-- resource minor=\"%u\"", minor);
+		if (resname)
+			printf(" name=\"%s\"", resname);
+		printf(" not available or not yet created -->\n");
+		return 0;
+	}
+	printf("<resource minor=\"%u\"", minor);
+	if (resname)
+		printf(" name=\"%s\"", resname);
+
+	if (state.conn == StandAlone && state.disk == Diskless) {
+		printf(" cs=\"Unconfigured\" />\n");
+		return 0;
+	}
+
+	printf( /* connection state */
+		" cs=\"%s\""
+		/* role */
+		" ro1=\"%s\" ro2=\"%s\""
+		/* disk state */
+		" ds1=\"%s\" ds2=\"%s\"",
+	       conns_to_name(state.conn),
+	       roles_to_name(state.role),
+	       roles_to_name(state.peer),
+	       disks_to_name(state.disk),
+	       disks_to_name(state.pdsk));
+
+	/* io suspended ? */
+	if (state.susp)
+		printf(" suspended");
+	/* reason why sync is paused */
+	if (state.aftr_isp)
+		printf(" aftr_isp");
+	if (state.peer_isp)
+		printf(" peer_isp");
+	if (state.user_isp)
+		printf(" user_isp");
+
+	if (consume_tag_int(T_sync_progress, rtl, &synced))
+		printf(" resynced_percent=\"%i.%i\"", synced / 10, synced % 10);
+
+	printf(" />\n");
+	return 0;
+}
+
+int sh_status_scmd(struct drbd_cmd *cm __attribute((unused)),
+		int minor, unsigned short *rtl)
+{
+/* variable prefix; maybe rather make that a command line parameter?
+ * or use "drbd_sh_status"? */
+#define _P ""
+	union drbd_state_t state = { .i = 0 };
+	int available = 0;
+	int synced = 0;
+
+	printf("%s_minor=%u\n", _P, minor);
+	printf("%s_res_name=%s\n", _P, shell_escape(resname ?: "UNKNOWN"));
+
+	available = consume_tag_int(T_state_i,rtl,(int*)&state.i);
+
+	if (state.conn == StandAlone && state.disk == Diskless) {
+		printf("%s_known=%s\n\n", _P,
+			available ? "Unconfigured"
+			          : "NA # not available or not yet created");
+		printf("%s_cstate=Unconfigured\n", _P);
+		printf("%s_role=\n", _P);
+		printf("%s_peer=\n", _P);
+		printf("%s_disk=\n", _P);
+		printf("%s_pdisk=\n", _P);
+		printf("%s_flags_susp=\n", _P);
+		printf("%s_flags_aftr_isp=\n", _P);
+		printf("%s_flags_peer_isp=\n", _P);
+		printf("%s_flags_user_isp=\n", _P);
+		printf("%s_resynced_percent=\n", _P);
+		return 0;
+	}
+
+	printf( "%s_known=Configured\n\n"
+		/* connection state */
+		"%s_cstate=%s\n"
+		/* role */
+		"%s_role=%s\n"
+		"%s_peer=%s\n"
+		/* disk state */
+		"%s_disk=%s\n"
+		"%s_pdsk=%s\n\n",
+		_P,
+		_P, conns_to_name(state.conn),
+		_P, roles_to_name(state.role),
+		_P, roles_to_name(state.peer),
+		_P, disks_to_name(state.disk),
+		_P, disks_to_name(state.pdsk));
+
+	/* io suspended ? */
+	printf("%s_flags_susp=%s\n", _P, state.susp ? "1" : "");
+	/* reason why sync is paused */
+	printf("%s_flags_aftr_isp=%s\n", _P, state.aftr_isp ? "1" : "");
+	printf("%s_flags_peer_isp=%s\n", _P, state.peer_isp ? "1" : "");
+	printf("%s_flags_user_isp=%s\n\n", _P, state.user_isp ? "1" : "");
+
+	printf("%s_resynced_percent=", _P);
+
+	if (consume_tag_int(T_sync_progress, rtl, &synced))
+		printf("%i.%i\n", synced / 10, synced % 10);
+	else
+		printf("\n");
+
+	printf("\n_sh_status_process\n\n\n");
+
+	fflush(stdout);
+	return 0;
+#undef _P
+}
+
+int role_scmd(struct drbd_cmd *cm __attribute((unused)),
 	       int minor __attribute((unused)),
 	       unsigned short *rtl)
 {
-	union drbd_state_t state;
+	union drbd_state_t state = { .i = 0 };
 	consume_tag_int(T_state_i,rtl,(int*)&state.i);
 	if ( state.conn == StandAlone &&
 	     state.disk == Diskless) {
@@ -1240,7 +1578,7 @@ int cstate_scmd(struct drbd_cmd *cm __attribute((unused)),
 		int minor __attribute((unused)),
 		unsigned short *rtl)
 {
-	union drbd_state_t state;
+	union drbd_state_t state = { .i = 0 };
 	consume_tag_int(T_state_i,rtl,(int*)&state.i);
 	if ( state.conn == StandAlone &&
 	     state.disk == Diskless) {
@@ -1255,7 +1593,7 @@ int dstate_scmd(struct drbd_cmd *cm __attribute((unused)),
 		int minor __attribute((unused)),
 		unsigned short *rtl)
 {
-	union drbd_state_t state;
+	union drbd_state_t state = { .i = 0 };
 	consume_tag_int(T_state_i,rtl,(int*)&state.i);
 	if ( state.conn == StandAlone &&
 	     state.disk == Diskless) {
@@ -1271,7 +1609,7 @@ int uuids_scmd(struct drbd_cmd *cm,
 	       unsigned short *rtl)
 {
 	__u64 *uuids;
-	int flags;
+	int flags = flags;
 	unsigned int len;
 
 	if(!consume_tag_blob(T_uuids,rtl,(char **) &uuids,&len)) {
@@ -1295,13 +1633,19 @@ int uuids_scmd(struct drbd_cmd *cm,
 	return 0;
 }
 
-static struct drbd_cmd *find_cmd_by_name(const char* name)
+static struct drbd_cmd *find_cmd_by_name(char *name)
 {
 	unsigned int i;
 
-	for(i=0;i<ARRY_SIZE(commands);i++) {
-		if(!strcmp(name,commands[i].cmd)) {
-			return commands+i;
+	if (!strcmp(name, "state")) {
+		fprintf(stderr, "'%s ... state' is deprecated, use '%s ... role' instead.\n",
+			cmdname, cmdname);
+		name = "role";
+	}
+
+	for (i = 0; i < ARRY_SIZE(commands); i++) {
+		if (!strcmp(name, commands[i].cmd)) {
+			return commands + i;
 		}
 	}
 	return NULL;
@@ -1448,7 +1792,7 @@ int print_broadcast_events(unsigned int seq, int u __attribute((unused)),
 	switch (reply->packet_type) {
 	case P_get_state:
 		if(consume_tag_int(T_state_i,reply->tag_list,(int*)&state.i)) {
-			printf("%u ST %d { cs:%s st:%s/%s ds:%s/%s %c%c%c%c }\n",
+			printf("%u ST %d { cs:%s ro:%s/%s ds:%s/%s %c%c%c%c }\n",
 			       seq,
 			       reply->minor,
 			       conns_to_name(state.conn),
@@ -1528,8 +1872,8 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 	struct drbd_nl_cfg_reply *reply;
 	struct drbd_tag_list *tl;
 	struct option *lo;
-	unsigned int seq=0;
-	int sk_nl,c,cont=1,rr,i,last;
+	unsigned int b_seq=0, r_seq=0;
+	int sk_nl,c,cont=1,rr = rr,i,last;
 	int unfiltered=0, all_devices=0;
 	int wfc_timeout=0, degr_wfc_timeout=0,timeout_ms;
 	struct timeval before,after;
@@ -1654,8 +1998,20 @@ int events_cmd(struct drbd_cmd *cm, int minor, int argc ,char **argv)
 
 		// dump_tag_list(reply->tag_list);
 
-		if(!unfiltered && cn_reply->seq <= seq) continue;
-		seq = cn_reply->seq;
+		/* There are two value spaces for sequence numbers. The first
+		   is the one created by this drbdsetup instance, the kernel's
+		   reply packets simply echo those sequence numbers.
+		   The second is created by the kernel's broadcast packets. */
+		if (!unfiltered) {
+			if (cn_reply->ack == 0) { // broadcasts
+				if (cn_reply->seq <= b_seq) continue;
+				b_seq = cn_reply->seq;
+			} else if (minor == reply->minor && cn_reply->ack == (__u32)getpid() + 1) {
+				// replies to drbdsetup packes and for this device.
+				if (cn_reply->seq <= r_seq) continue;
+				r_seq = cn_reply->seq;
+			}
+		}
 
 		if( all_devices || minor == reply->minor ) {
 			cont=cm->ep.proc_event(cn_reply->seq, wasb, reply);
@@ -1712,6 +2068,14 @@ int string_opt_usage(struct drbd_option *option, char* str, int strlen)
 			option->name, option->short_name);
 }
 
+void af_opt_xml(struct drbd_option *option)
+{
+	printf("\t<option name=\"%s\" type=\"addrfamily\">\n",option->name);
+	printf("\t\t<addrfamily>%s</addrfamily>\n", "IPv4");
+	printf("\t\t<addrfamily>%s</addrfamily>\n", "SCI");
+	printf("\t</option>\n");
+}
+
 void numeric_opt_xml(struct drbd_option *option)
 {
 	printf("\t<option name=\"%s\" type=\"numeric\">\n",option->name);
@@ -1917,7 +2281,7 @@ int open_cn()
 	}
 
 	my_nla.nl_family = AF_NETLINK;
-	my_nla.nl_groups = -1; //CN_IDX_DRBD;
+	my_nla.nl_groups = -1; //cn_idx
 	my_nla.nl_pid = getpid();
 
 	err = bind(sk_nl, (struct sockaddr *)&my_nla, sizeof(my_nla));
@@ -1953,9 +2317,9 @@ void prepare_nl_header(struct nlmsghdr* nl_hdr, int size)
 	nl_hdr->nlmsg_pid = getpid();
 	/* fill the connector header */
 	cn_hdr->id.val = CN_VAL_DRBD;
-	cn_hdr->id.idx = CN_IDX_DRBD;
+	cn_hdr->id.idx = cn_idx;
 	cn_hdr->seq = cn_seq++;
-	get_random_bytes(&cn_hdr->ack,sizeof(cn_hdr->ack));
+	cn_hdr->ack = getpid();
 	cn_hdr->len = size - sizeof(struct nlmsghdr) - sizeof(struct cn_msg);
 }
 
@@ -2048,28 +2412,32 @@ void close_cn(int sk_nl)
 
 void ensure_drbd_driver_is_present(void)
 {
-	struct drbd_tag_list *tl;
-	char buffer[4096];
-	int sk_nl, rr;
+	struct stat sb;
+	FILE *cn_idx_file;
+	int err;
 
-	sk_nl = open_cn();
-	/* Might print:
-	   Missing privileges? You should run this as root.
-	   Connector module not loaded? try 'modprobe cn'. */
-	if (sk_nl < 0) exit(20);
+	cn_idx = CN_IDX_DRBD;
+	cn_idx_file = fopen("/sys/module/drbd/parameters/cn_idx", "r");
+	if (cn_idx_file) {
+		unsigned int idx; /* gcc is picky */
+		if (fscanf(cn_idx_file, "%u", &idx))
+			cn_idx = idx;
+		fclose(cn_idx_file);
+	}
 
-	tl = create_tag_list(2);
-	add_tag(tl, TT_END, NULL, 0); // close the tag list
+	err = stat("/proc/drbd", &sb);
+	if (!err)
+		return;
 
-	tl->drbd_p_header->packet_type = P_get_state;
-	tl->drbd_p_header->drbd_minor = 0;
-	tl->drbd_p_header->flags = 0;
+	if (err == ENOENT)
+		fprintf(stderr, "DRBD driver appears to be missing\n");
+	else
+		fprintf(stderr, "Could not stat(\"/proc/drbd\"): %m\n");
 
-	rr = call_drbd(sk_nl, tl, (struct nlmsghdr*)buffer, 4096, 500);
-	/* Might print: (after 500ms)
-	   No response from the DRBD driver! Is the module loaded? */
-	close_cn(sk_nl);
-	if (rr == -2) exit(20);
+	fprintf(stderr, "do you need to load the module?\n"
+			"try: modprobe drbd\n");
+
+	exit(20);
 }
 
 int main(int argc, char** argv)
@@ -2078,9 +2446,14 @@ int main(int argc, char** argv)
 	struct drbd_cmd *cmd;
 	int rv=0;
 
-	chdir("/");
+	if (chdir("/")) {
+		/* highliy unlikely, but gcc is picky */
+		perror("cannot chdir /");
+		return -111;
+	}
 
-	if ( (cmdname = strrchr(argv[0],'/')) )
+	cmdname = strrchr(argv[0],'/');
+	if (cmdname)
 		argv[0] = ++cmdname;
 	else
 		cmdname = argv[0];
@@ -2098,6 +2471,7 @@ int main(int argc, char** argv)
 	/* it is enough to set it, value is ignored */
 	if (getenv("DRBD_DEBUG_DUMP_ARGV"))
 		debug_dump_argv = 1;
+	resname = getenv("DRBD_RESOURCE");
 
 	if (argc > 1 && (!strcmp(argv[1],"xml"))) {
 		if(argc >= 3) {
